"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isPropObj = exports.createRuntimeBindings = void 0;
var vue_1 = require("vue");
var utils_1 = require("./utils");
var createRuntimeBindings = function (_createRuntime // arg only used to infer Runtime type
) {
    var useSharedState = function () {
        var props = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            props[_i] = arguments[_i];
        }
        var cleanups = [];
        var runtime = (0, vue_1.inject)("runtime");
        var vueState = {};
        if (!runtime) {
            throw new Error("You must `provide` a `runtime` to the app at the root of the component tree");
        }
        props.forEach(function (prop) {
            var propName;
            var deep = false;
            if (typeof prop === "string") {
                propName = prop;
            }
            else if ((0, exports.isPropObj)(prop)) {
                propName = prop.prop;
                if (prop.loader) {
                    runtime.load(propName, prop.loader);
                }
                deep = !!prop.deep;
            }
            else {
                throw new Error("prop must be either a string or an object with the key prop");
            }
            // It uses a `ref` for each state prop instead of `reactive(vueState)` so state can be destructured
            vueState[propName] = (0, vue_1.ref)(runtime.state[propName]);
            if (deep || (0, utils_1.isPrimitive)(runtime.state[propName])) {
                // ref( runtime.state[propName] ) adds a proxy recursively to each property of runtime.state[propName]
                // so any property in any level of runtime.state[propName] will change when that property changes in vueState[propName].
                // However, if the change is not at the root of runtime.state[propName] then runtime.state subscribers won't be called
                cleanups.push(
                // watch returns an unwatch function
                (0, vue_1.watch)(vueState[propName], function syncState(newValue) {
                    var rawValue = (0, vue_1.toRaw)(newValue);
                    // If runtime.state[propName] === rawValue then runtime doesn't calls its subscribers
                    if (rawValue === runtime.state[propName] && deep) {
                        // newValue was mutated, meaning a deep value changed in newValue but the reference to newValue didn't.
                        // Therefore, we need to change the reference to newValue so that runtime is aware of the change
                        vueState[propName].value = (0, utils_1.shallowCopy)(rawValue);
                        // Assigning vueState[propName].value triggers the syncState watch function again,
                        // so we don't set runtime.state in this branch to avoid an infinite loop
                    }
                    else {
                        // Update runtime state with the new value so runtime subscribers are called
                        runtime.state[propName] = rawValue;
                    }
                }, { deep: deep }));
            }
            cleanups.push(
            // Runtime returns a unsubscribe function
            runtime.subscribe(propName, function (value) {
                // If vueState[propName].value === toRaw(value) then Vue bails out of setting the value
                vueState[propName].value = value;
            }));
        });
        // We call all the clean-up functions when the composable unmounts
        (0, vue_1.onUnmounted)(function () { return cleanups.forEach(function (cleanup) { return cleanup(); }); });
        return vueState;
    };
    return {
        useSharedState: useSharedState,
    };
};
exports.createRuntimeBindings = createRuntimeBindings;
var isPropObj = function (prop) { return !!prop.prop; };
exports.isPropObj = isPropObj;
